import Foundation
import TSCBasic
import TSCUtility
import TuistAutomation
import TuistCore
import TuistGenerator
import TuistGraph

protocol WorkspaceMapperFactorying {
    /// Returns the default workspace mapper.
    /// - Returns: A workspace mapping instance.
    func `default`(config: Config) -> [WorkspaceMapping]

    /// Returns a mapper to generate cacheable prorjects.
    /// - Parameter config: The project configuration.
    /// - Parameter includedTargets: The list of targets to cache.
    /// - Returns: A workspace mapping instance.
    func cache(config: Config, includedTargets: Set<String>) -> [WorkspaceMapping]

    /// Returns a mapper for automation commands like build and test.
    /// - Parameter config: The project configuration.
    /// - Parameter workspaceDirectory: The directory where the workspace will be generated.
    /// - Returns: A workspace mapping instance.
    func automation(config: Config, workspaceDirectory: AbsolutePath) -> [WorkspaceMapping]
}

final class WorkspaceMapperFactory: WorkspaceMapperFactorying {
    private let projectMapper: ProjectMapping

    init(projectMapper: ProjectMapping) {
        self.projectMapper = projectMapper
    }

    func cache(config: Config, includedTargets: Set<String>) -> [WorkspaceMapping] {
        var mappers = self.default(config: config)
        mappers += [GenerateCacheableSchemesWorkspaceMapper(includedTargets: includedTargets)]
        return mappers
    }

    func automation(config: Config, workspaceDirectory: AbsolutePath) -> [WorkspaceMapping] {
        var mappers: [WorkspaceMapping] = []
        mappers.append(AutomationPathWorkspaceMapper(workspaceDirectory: workspaceDirectory))
        mappers += self.default(config: config)
        if config.generationOptions.contains(.autogenerationOptions(.disabled)) {
            mappers.append(
                AutogeneratedProjectSchemeWorkspaceMapper(
                    codeCoverageMode: config.codeCoverageMode,
                    testingOptions: []
                )
            )
        }

        return mappers
    }

    func `default`(config: Config) -> [WorkspaceMapping] {
        var mappers: [WorkspaceMapping] = []

        mappers.append(
            ProjectWorkspaceMapper(mapper: projectMapper)
        )

        mappers.append(
            TuistWorkspaceIdentifierMapper()
        )

        mappers.append(
            IDETemplateMacrosMapper()
        )

        // Auto-generation of project scheme
        if let testingOptions = config.autogenerationTestingOptions {
            mappers.append(
                AutogeneratedProjectSchemeWorkspaceMapper(
                    codeCoverageMode: config.codeCoverageMode,
                    testingOptions: testingOptions
                )
            )
        }

        mappers.append(
            ModuleMapMapper()
        )

        if let lastUpgradeVersion = lastUpgradeCheckOverride(config: config) {
            mappers.append(
                LastUpgradeVersionWorkspaceMapper(
                    lastUpgradeVersion: lastUpgradeVersion
                )
            )
        }

        return mappers
    }

    // MARK: - Fileprivate

    fileprivate func lastUpgradeCheckOverride(config: Config) -> Version? {
        config.generationOptions.compactMap { item -> Version? in
            switch item {
            case let .lastUpgradeCheck(version):
                return version
            default:
                return nil
            }
        }.first
    }
}
